"""Tools for setting permissions on commands."""
from typing import Any, Optional

from nextcord.enums import Enum

from .commands import SlashCommandSubGroup, SlashSubCommand, TopLevelCommand


class ApplicationCommandPermissionType(Enum):
    """Enum representing the permission type of an application command."""

    role = 1
    user = 2


UserPermissions = dict[int, bool]
# More accurately, list[nextcord.types.interactions.ApplicationCommandPermissions],
# but that just causes type checker errors.
ApiPermissions = list[dict[str, int]]


def _to_dict(ids: tuple[int, ...], value: bool) -> UserPermissions:
    """Create a dict with the same value for every key."""
    return {id: value for id in ids}


def _dump_permissions(
    *, roles: Optional[UserPermissions] = None, users: Optional[UserPermissions] = None
) -> ApiPermissions:
    """Dump permissions to the JSON structure accepted by the API."""
    return [
        *(
            {
                "id": role_id,
                "type": ApplicationCommandPermissionType.role.value,
                "permission": value,
            }
            for role_id, value in (roles or {}).items()
        ),
        *(
            {
                "id": user_id,
                "type": ApplicationCommandPermissionType.user.value,
                "permission": value,
            }
            for user_id, value in (users or {}).items()
        ),
    ]


def _update_permissions(dest: dict[int, ApiPermissions], source: dict[int, ApiPermissions]):
    """Update permissions for a command, extending rather than overwriting."""
    for guild_id, permissions in source.items():
        if guild_id in dest:
            dest[guild_id].extend(permissions)
        else:
            dest[guild_id] = permissions


class PermissionsSetter:
    """Decorator to set permissions on a command."""

    def __call__(self, command: TopLevelCommand) -> TopLevelCommand:
        """Set permissions on a command."""
        if not isinstance(command, TopLevelCommand):
            raise self._wrong_type(command)
        self._apply_permissions(command)
        return command

    def _wrong_type(self, command: Any) -> TypeError:
        """Construct an error when the given command is of the wrong type."""
        if isinstance(command, SlashCommandSubGroup):
            attempt = "subcommand group."
        elif isinstance(command, SlashSubCommand):
            attempt = "subcommand."
        elif callable(command):
            attempt = (
                "plain function. Make sure any permissions decorators are "
                "before the command decorator."
            )
        else:
            attempt = repr(type(command))
        return TypeError("Permissions can only be set on a top level command, not a " f"{attempt}")

    def _apply_permissions(self, command: TopLevelCommand):
        """Apply permissions to a command."""
        raise NotImplementedError


class AllPermissionsSetter(PermissionsSetter):
    """Decorator to set permissions on a command for all guilds."""

    def __init__(self, roles: dict[int, UserPermissions], users: dict[int, UserPermissions]):
        """Store the permissions."""
        self.permissions = {}
        _update_permissions(
            self.permissions,
            {
                guild_id: _dump_permissions(roles=role_permissions)
                for guild_id, role_permissions in roles.items()
            },
        )
        _update_permissions(
            self.permissions,
            {
                guild_id: _dump_permissions(users=user_permissions)
                for guild_id, user_permissions in users.items()
            },
        )

    def _apply_permissions(self, command: TopLevelCommand):
        """Set permissions on a command."""
        _update_permissions(command.permissions, self.permissions)  # type: ignore


class GuildPermissionsSetter(PermissionsSetter):
    """Decorator to set permissions on a command for one guild."""

    def __init__(self, roles: UserPermissions, users: UserPermissions, guild_id: Optional[int]):
        """Store the permissions."""
        self.permissions = _dump_permissions(roles=roles, users=users)
        self.guild_id = guild_id

    def _apply_permissions(self, command: TopLevelCommand):
        """Set permissions on a command."""
        guild_id = self.guild_id or command.guild_id
        if not guild_id:
            raise TypeError(
                f"Command {command.name!r} is not for a specific guild, so a "
                "guild ID must be set for permissions."
            )
        _update_permissions(command.permissions, {guild_id: self.permissions})  # type: ignore


def global_permissions(
    roles: dict[int, UserPermissions], users: dict[int, UserPermissions]
) -> AllPermissionsSetter:
    """Create a decorator to set permissions for multiple guilds."""
    return AllPermissionsSetter(roles, users)


def guild_permissions(
    roles: Optional[UserPermissions] = None,
    users: Optional[UserPermissions] = None,
    guild_id: Optional[int] = None,
) -> GuildPermissionsSetter:
    """Create a decorator to set permissions for a guild."""
    return GuildPermissionsSetter(roles or {}, users or {}, guild_id)


def allow_roles(*role_ids: int, guild_id: Optional[int] = None) -> GuildPermissionsSetter:
    """Allow certain roles to use a command in a guild."""
    return guild_permissions(roles=_to_dict(role_ids, True), guild_id=guild_id)


def disallow_roles(*role_ids: int, guild_id: Optional[int] = None) -> GuildPermissionsSetter:
    """Prevents certain roles from using a command in a guild."""
    return guild_permissions(roles=_to_dict(role_ids, False), guild_id=guild_id)


def allow_users(*user_ids: int, guild_id: Optional[int] = None) -> GuildPermissionsSetter:
    """Allow certain users to use a command in a guild."""
    return guild_permissions(users=_to_dict(user_ids, True), guild_id=guild_id)


def disallow_users(*user_ids: int, guild_id: Optional[int] = None) -> GuildPermissionsSetter:
    """Prevents certain users from using a command in a guild."""
    return guild_permissions(users=_to_dict(user_ids, False), guild_id=guild_id)



"""Interface for specifying option choices."""
from __future__ import annotations

import functools
from typing import Any, Generic, Optional, Type, TypeVar, Union

ChoiceType = Union[str, int, float]
CT = TypeVar("CT")


class Choice(Generic[CT]):
    """A single choice as part of an option's choices."""

    name: str
    value: CT

    def __init__(self, name: str, value: CT):
        """Store the choice attributes."""
        self.name = name
        self.value = value

    def __repr__(self) -> str:
        """Return the string representation of the choice."""
        return f"Choice(name={self.name!r}, value={self.value!r})"


class Choices(Generic[CT]):
    """A class containing each of the choices for some option.

    Example usage:

    ```python
    class IcecreamFlavors(Choices):
        vanilla = "Vanilla"
        chocolate = "Chocolate"
        strawberry = "Strawberry"
    ```

    Or for integer options:

    ```python
    class Rating(Choices):
        one = Choice("Very bad", 1)
        two = Choice("Bad", 2)
        three = Choice("OK", 3)
        four = Choice("Good", 4)
        five = Choice("Great", 5)
    ```
    """

    name: str
    value: CT

    @classmethod
    def _get_attr_choice(
        cls, name: str, value: Any
    ) -> Optional[tuple[Choice[ChoiceType], Type[ChoiceType]]]:
        """Convert an attribute to a choice."""
        if isinstance(value, str):
            # 'name' is the name of the Python attribute, which should be the
            # internally-used *value* for the choice, and 'value' is the value
            # of the Python attribute, which should be the user-displayed
            # *name* of the choice.
            return Choice(name=value, value=name), str
        elif isinstance(value, Choice):
            return value, type(value.value)
        return None

    @classmethod
    @functools.cache
    def _get_choices(cls) -> tuple[list[Choice[CT]], Type[CT]]:
        """Get the JSON data for registering these choices with the API."""
        choices = []
        choice_type = None
        for name, value in cls.__dict__.items():
            if name.startswith("_"):
                continue
            if not (choice_and_type := cls._get_attr_choice(name, value)):
                continue
            choice, this_type = choice_and_type
            if choice_type is None:
                choice_type = this_type
            elif choice_type != this_type:
                raise TypeError(
                    f"All choices must be of the same type, but {name} is "
                    f"{this_type} and previous choices were {choice_type}."
                )
            choices.append(choice)
        if choice_type is None:
            raise ValueError("There must be at least one choice.")
        return choices, choice_type  # type: ignore

    @classmethod
    @functools.cache
    def _find_choice(cls, value: CT) -> Choices[CT]:
        """Find the choice with the given value."""
        choices, _choice_type = cls._get_choices()
        for choice in choices:
            if choice.value == value:
                return cls(choice.name, choice.value)
        raise ValueError(f"Recieved {cls.__name__!r} choice with invalid value {value!r}.")

    def __init__(self, name: str, value: CT):
        """Store the attributes of a single choice."""
        self.name = name
        self.value = value

    def __repr__(self) -> str:
        """Return the string representation of the choice."""
        return f"{self.__class__.__name__}(name={self.name!r}, value={self.value!r})"



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dslash']

package_data = \
{'': ['*']}

install_requires = \
['docstring-parser>=0.12,<0.13', 'nextcord>=2.0.0-alpha.3,<3.0.0']

setup_kwargs = {
    'name': 'dslash',
    'version': '0.3.3',
    'description': 'A library which supplements Nextcord by adding support for slash commands.',
    'long_description': '# DSlash\n\n![Version: 0.3.3](https://img.shields.io/badge/Version-0.3.3-red?style=flat-square)\n[![Code Style: black](https://img.shields.io/badge/Code%20Style-black-black?style=flat-square)](https://github.com/psf/black)\n[![License: MIT](https://img.shields.io/badge/License-MIT-orange?style=flat-square)](./LICENSE)\n[![PyPI: dslash](https://img.shields.io/badge/PyPI-dslash-green?style=flat-square)](https://pypi.org/project/dslash)\n![Python: ^3.9](https://img.shields.io/badge/python-%5E3.9-blue?style=flat-square)\n\nA library which supplements [Nextcord](https://github.com/nextcord/nextcord)\n(a fork of Discord.py) by adding support for slash commands.\n\nDocumentation is still a work in progress, and the library should currently be\nconsidered unstable.\n\nYou can install it using pip, eg. `pip install dslash`.\n\n## Example\n\n```python\nimport random\nimport logging\nimport traceback\nimport typing\n\nfrom nextcord import Embed, Interaction, Member, Role\nfrom dslash import Choices, CommandClient, SlashCommandInvokeError, allow_roles, option\n\n\nGUILD_ID = ...\nADMIN_ROLE_ID = ...\nTOKEN = ...\n\nlogging.basicConfig(level=logging.INFO)\nclient = CommandClient(guild_id=GUILD_ID)\n\n\n@client.event\nasync def on_ready():\n    print(f\'Logged in as {client.user}.\')\n\n\n@client.command()\nasync def roll(interaction: Interaction, sides: typing.Optional[int]):\n    """Roll a dice.\n\n    :param sides: How many sides (default 6).\n    """\n    value = random.randint(1, sides or 6)\n    await interaction.response.send_message(f\'You got: {value}\')\n\n\nimages = client.group(\'images\', \'Cute image commands.\')\n\n\n@images.subcommand()\nasync def cat(interaction: Interaction):\n    """Get a cat image."""\n    await interaction.response.send_message(\n        embed=Embed().set_image(url=\'https://cataas.com/cat\')\n    )\n\n\n@images.subcommand()\nasync def dog(interaction: Interaction):\n    """Get a dog image."""\n    await interaction.response.send_message(\n        embed=Embed().set_image(url=\'https://placedog.net/500?random\')\n    )\n\n\n@images.subcommand(name=\'any\')\nasync def any_(interaction: Interaction):\n    """Get any random image."""\n    await interaction.response.send_message(\n        embed=Embed().set_image(url=\'https://picsum.photos/600\')\n    )\n\n\nadmin = client.group(\n    \'admin\',\n    \'Admin-only commands.\',\n    default_permission=False,\n    permissions=allow_roles(ADMIN_ROLE_ID)\n)\nroles = admin.subgroup(\'roles\', \'Commands to manage roles.\')\n\n\n@roles.subcommand(name=\'del\')\nasync def del_(interaction: Interaction, role: Role):\n    """Delete a role.\n\n    :param role: The role to delete.\n    """\n    await role.delete()\n    await interaction.response.send_message(\'Deleted the role.\', ephemeral=True)\n\n\n@allow_roles(ADMIN_ROLE_ID)\n@client.command(default_permission=False)\nasync def ban(interaction: Interaction, user: Member):\n    """Ban a user.\n\n    :param user: The user to ban.\n    """\n    await user.ban()\n    await interaction.response.send_message(\'Banned the user.\', ephemeral=True)\n\n\nclass RPSChoices(Choices):\n    rock = \'Rock\'\n    paper = \'Paper\'\n    scissors = \'Scissors\'\n    gun = \'Gun\'\n\n\n@client.command()\nasync def rps(interaction: Interaction, choice: RPSChoices):\n    """Play rock, paper, scissors.\n\n    :param choice: Your choice.\n    """\n    if choice == RPSChoices.gun:\n        await interaction.response.send_message("That\'s cheating!")\n    else:\n        await interaction.response.send_message(f\'You picked {choice.name}.\')\n\n\nclient.run(TOKEN)\n```\n\n## Planned Features\n\n- Class-based command groups, like `nextcord.ext.commands` cogs.\n\nCompatibility with `nextcord.ext.commands` is not planned.\n\n## Development\n\nAs well as Python 3.9+, this project requires Poetry for development.\n[Click this link for installation instructions](https://python-poetry.org/docs/master/#installation),\nor:\n\n- #### \\*nix (Linux/MacOS)\n\n  `curl -sSL https://raw.githubusercontent.com/python-poetry/poetry/master/install-poetry.py | python -`\n\n- #### Windows Powershell\n\n  `(Invoke-WebRequest -Uri https://raw.githubusercontent.com/python-poetry/poetry/master/install-poetry.py -UseBasicParsing).Content | python -`\n\nOnce you have Poetry installed:\n\n1. **Create a virtual environment:** `poetry shell`\n2. **Install dependencies:** `poetry install`\n\nThe following commands are then available:\n\n- `poe format` - Run auto-formatting and linting.\n\nPrefix these with `poetry run` if outside of the Poetry shell.\n',
    'author': 'Artemis',
    'author_email': 'me@arty.li',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/artemis21/dslash',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
